# Code to handle the RoutePlanner RPL3 database format
# (Designed to be easily parsable in Python and humanly grokable too)
#
# Copyright 1996-98 Chris Lawrence
# This file may be freely distributed under the terms of the RoutePlanner
# license.  A copy should appear as 'LICENSE' in the archive that this
# file was included in.
#
# db = rpdbase.RPDatabase(filename) - Load a database
# db.Open(filename) - Replace the database by loading another
# db.Save(filename) - Write the database
#
# db.cities - The list of city instances
# db.routes - The list of route instances

import rpcity, rproute, string, time, os, re, sys

VERID = "##VERSION##"
VERSION = "RoutePlanner ##VERSION##" 
COPYRIGHT = '(C) 1994-2000 Chris Lawrence'

invalidformat = 'rpdbase.invalidformat'

LEAST_DISTANCE = 0
LEAST_TIME     = 1
PREFERRED      = 2

methods = ['Shortest distance', 'Shortest time', 'Preferred']

UNITS_METRIC   = 'km'
UNITS_IMPERIAL = 'mi'

units = { 'Metric (km, l)' : UNITS_METRIC,
          'Imperial (mi, gal)' : UNITS_IMPERIAL }

MAXDIST = 2**24

def ConvertDistance(length, units, newunits):
    if units == newunits:
        return float(length)
    if newunits == UNITS_METRIC:
        return length * 1.6093
    else:
        return length / 1.6093

def ConvertToTime(speedcode, speed, distance, runits, cunits, codetable):
    if speedcode == 0:
        return ( 60 * distance / speed )
    elif speedcode == 1:
        return speed
    else:
        return ( 60 * ConvertDistance(distance, runits, cunits) /
                 codetable[speedcode] )

def Efficiency(distance, dunits, tunits, uunits, speedcode, mpg, defmpg):
    distance = ConvertDistance(distance, dunits, tunits)
    if speedcode >= 2:
        our_mpg = mpg[speedcode]
    else:
        our_mpg = defmpg
        
    if uunits != tunits:
        if uunits == UNITS_METRIC:
            our_mpg = our_mpg * 1.609/3.78
        else:
            our_mpg = our_mpg * 3.78/1.609
    try:
        return distance / our_mpg
    except ZeroDivisionError, text:
        print 'Warning: MPG is 0 for speedcode %d: %s' % (speedcode, text)
        return 0

def Least_Distance(distance, speed, speedcode, runits, tunits, codetable,
                   prefertable, defaultprefer):
    return ConvertDistance(distance, runits, UNITS_METRIC)

def Least_Time(distance, speed, speedcode, runits, tunits, codetable,
               prefertable, defaultprefer):
    return ConvertToTime(speedcode, speed, distance, runits, tunits,
                         codetable)

def Preferred(distance, speed, speedcode, runits, tunits, codetable,
              prefertable, defaultprefer):
    dist = ConvertDistance(distance, runits, UNITS_METRIC)
    if speedcode >= 2:
        return (dist * 101) / (prefertable[speedcode] + 1)
    else:
        return (dist * 101) / (defaultprefer + 1)

def IsSameRoute(route1, route2):
    if route1 == route2: return 1
    if '/' in route1: return 0
    if not '/' in route2: return 0

    bits = string.split(route1, ' ', 1)
    usedchar = ' '
    if len(bits) == 1:
        bits = string.split(route1, '-', 1)
        usedchar = '-'
        if len(bits) == 1:
            return 0

    # Ok, we're looking for 'BIT-n/n/BIT/n/n' or 'BIT a/BIT/b/c'
    # or the like.
    routechars = r"[\w\s]"
    rexp = r"^%s[%s](%s+/)*%s(/%s+)*$" % (bits[0], usedchar, routechars,
                                          bits[1], routechars)
    return re.search(rexp, route2)

def citycmp(city1, city2):
    return cmp(string.lower(str(city1)), string.lower(str(city2)))

def routecmp(route1, route2):
    return citycmp(route1.city[0], route2.city[0]) or \
           citycmp(route1.city[1], route2.city[1]) or \
           cmp(route1.distance, route2.distance) or \
           cmp(string.lower(route1.name), string.lower(route2.name))

class RPDatabase:
    def __init__(self, filename=None, quiet=0, progressbar=None):
        if filename:
            self.Open(filename, quiet, progressbar)
        else:
            self.cities = self.routes = []

    def Open(self, filename, quiet=0, progressbar=None):
        fh = open(filename, 'r', -1)
        stuff = fh.readline()
        if stuff[0:4] != 'RPL3':
            raise invalidformat, filename+' is not a valid RPL3 file'
        self.comment = stuff[5:-1]

        cities = int(fh.readline()[:-1])
        citylist = [None]*cities
        cityhash = {}
        if not quiet:
            sys.stdout.write('\nCities: ')
            sys.stdout.flush()

        if progressbar:
            progressbar.set_max(cities)
            progressbar.set_label('Loading Cities')
            progressbar.set_current(0)

        for i in xrange(cities):
            data = fh.readline()[:-1]
            fields = string.split(data, '|')
            x = rpcity.City( fields )
            cityhash[ fields[0] ] = (x, i)
            cityhash[ string.lower(fields[0]) ] = (x, i)
            citylist[ i ] = x
            if not (i % 100):
                if not quiet:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                if progressbar:
                    progressbar.set_current(i)

        routes = int(fh.readline()[:-1])
        routelist = [None]*routes
        if not quiet:
            sys.stdout.write('\nRoutes: ')
            sys.stdout.flush()

        if progressbar:
            progressbar.set_max(routes)
            progressbar.set_label('Loading Routes')
            progressbar.set_current(0)

        for i in xrange(routes):
            data = fh.readline()[:-1]
            fields = string.split(data, '|')
            routelist[ i ] = rproute.Route( fields, cityhash )
            if not (i % 100):
                if not quiet:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                if progressbar:
                    progressbar.set_current(i)
        
        if not quiet:
            sys.stdout.write('\n')
            sys.stdout.flush()

        self.cities = citylist
        # self.cities.sort(lambda x, y: cmp(
        #                  string.lower(str(x)), string.lower(str(y)) ))

        self.filename = filename
        self.cityhash = cityhash
        self.routes = routelist
        self.author = fh.readline()[:-1]
        self.author_email = fh.readline()[:-1]

        # The following are raw data stored in the file as defaults that
        # can be overridden by the user (using a prefs file)
        self.units  = fh.readline()[:-1]
        self.classifications = eval(fh.readline()[:-1])
        self.speed  = eval(fh.readline()[:-1])
        self.mpg    = eval(fh.readline()[:-1])
        self.prefer = eval(fh.readline()[:-1])
        self.defmpg = eval(fh.readline()[:-1])
        self.defprf = eval(fh.readline()[:-1])
        fh.close()

    def Sort(self):
        self.cities.sort(citycmp)
        self.routes.sort(routecmp)
        
    def Save(self, filename, comment=None):
        if os.path.exists(filename):
            os.rename(filename, filename+'~')

        if not comment:
            comment = self.comment

        self.Sort()
        fh = open(filename, 'w', -1)
        fh.write('RPL3 %s\n' % comment)

        fh.write( '%d\n' % len(self.cities) )
        for city in self.cities:
            fh.write( `city`+'\n' )

        fh.write( '%d\n' % len(self.routes) )
        for route in self.routes:
            fh.write( `route`+'\n' )

        fh.write( self.author + '\n' + self.author_email + '\n' )
        fh.write( self.units + '\n' + `self.classifications` + '\n' +
                  `self.speed` + '\n' + `self.mpg` + '\n' +
                  `self.prefer` + '\n' + `self.defmpg` + '\n' +
                  `self.defprf` + '\n' )
        fh.close()

    def CitiesMatching(self, pat):
        return map(lambda x, y=self: y.cities[x], self.CitiesMatchIndex(pat))

    def CitiesMatchIndex(self, pat, visible=()):
        x = self.cityhash.get(pat)
        if x:
            return [x[1]]       
        
        matching = []
        
        if ',' in pat:
            city, state = re.split('\s*,\s*', pat, 1)
            exp = re.compile(re.escape(city)+'[^,]*, '+re.escape(state), re.I)
        else:
            exp = re.compile(re.escape(pat), re.I)
        
        if not visible:
            visible = range(len(self.cities))

        for i in visible:
            if exp.match(str(self.cities[i])) is not None:
                matching.append(i)

        return matching
    
    # The following routine is adapted from a C algorithm created by
    # Jim Butterfield and placed in the public domain.
    #
    # It is apparently based on Dijkstra's algorithm for shortest
    # paths; it is O(n^2)
    def OldNavigate(self, start, end, method=LEAST_DISTANCE, units=None,
                    codetable=None, prefertable=None, defaultprefer = None):
        if not units:         units = self.units
        if not codetable:     codetable = self.speed
        if not prefertable:   prefertable = self.prefer
        if not defaultprefer: defaultprefer = self.defprf
        
        # Fake an enumeration
        WORKLIST, NEXTCITY, MINDIST = 0, 1, 2

        citydict = {}
        weightings = {}

        for city in self.cities:
            citydict[id(city)] = [None, None, MAXDIST]

        elink = end
        minival = 0

        usefunc = [Least_Distance, Least_Time, Preferred]
        methodfunc = usefunc[method]

        citydict[ id(end) ] = [None, None, 0]
        thisdist, searchcity = MAXDIST, end

        startstr = id(start)
        startcityinf = citydict[startstr]

        while searchcity:
            searchstr = id(searchcity)
            searchcityinf = citydict[searchstr]

            if (searchcityinf[MINDIST] < startcityinf[MINDIST]):
                minival = searchcityinf[MINDIST]
                #print 'At %s, distance %d' % (searchstr, minival)

                for (nextroute, nextcol, othercity) in searchcity.roads:
                    rtestr = repr(nextroute)
                    dist = weightings.get(rtestr)
                    if not dist:
                        dist = methodfunc(nextroute.distance, nextroute.speed,
                                          nextroute.speedcode, nextroute.units,
                                          units, codetable, prefertable,
                                          defaultprefer)
                        weightings[rtestr] = dist

                    othercitystr = id(othercity)
                    othercityinf = citydict[othercitystr]
                    
                    travel = minival + dist

                    # New shortest distance
                    if othercityinf[MINDIST] > travel:
                        othercityinf[MINDIST] = travel
                        othercityinf[NEXTCITY] = searchcity

                        if (not othercityinf[WORKLIST] and
                            elink is not othercity):
                            (blink, flink) = (searchcity,
                                              searchcityinf[WORKLIST])
                            
                            if flink:
                                flinkstr = id(flink)
                                flinkinf = citydict[flinkstr]

                            while flink and travel > flinkinf[MINDIST]:
                                (blink, flink) = (flink, flinkinf[WORKLIST])
                                if flink:
                                    flinkstr = id(flink)
                                    flinkinf = citydict[flinkstr]

                            blinkstr = id(blink)
                            blinkinf = citydict[blinkstr]
                            blinkinf[WORKLIST] = othercity
                            othercityinf[WORKLIST] = flink
                            
                            if not flink:
                                elink = othercity

                    thisdist = min(travel, thisdist)

            searchcity = searchcityinf[WORKLIST]
            searchcityinf[WORKLIST] = None
        
        trail = []
        searchcity, flink = start, startcityinf[NEXTCITY]
        while flink:
            for i in xrange( len(searchcity.roads) ):
                if (searchcity.roads[i][0].city[1-searchcity.roads[i][1]]
                    is flink):
                    route = (searchcity.roads[i][0], 1-searchcity.roads[i][1])
            trail.append( route )
            searchcity, flink = flink, citydict[id(flink)][NEXTCITY]

        return trail

    # New navigation routine; based on example by
    # Aaron Watters <aaron@cs.rutgers.edu> posted to comp.lang.python 
    # 12/22/1997
    #
    # His example used his kjbuckets C extension for graphs; I'm using
    # another representation (they benchmarked the same).  kjSet seems
    # to buy some performance over a straight dictionary, so it is used
    # if available.
    #
    # Interestingly enough, PQEquivMax seems to be faster than PQueue;
    # no idea why...
    #
    # Adapted to use the PQueue extension class by Andrew Snare
    # on 1/3/2000
    def NewNavigate(self, start, end, method=LEAST_DISTANCE, units=None,
                    codetable=None, prefertable=None, defaultprefer=None):
        if start==end:
            return []

        if not units:         units = self.units
        if not codetable:     codetable = self.speed
        if not prefertable:   prefertable = self.prefer
        if not defaultprefer: defaultprefer = self.defprf
        
        usefunc = [Least_Distance, Least_Time, Preferred]
        methodfunc = usefunc[method]
        weightings = {}

        from pqueue import PQueue
        Q = PQueue()
        
        reverse_path = {start : start}
        costs = {start : 0}
        try:
            # from space import hosieries
            from kjbuckets import kjSet
            seen = kjSet()
        except:
            seen = {} # Emulate a kjSet

        for neighbor in start.roads:
            pair = start, neighbor[2]
            wt = weightings.get(pair)
            if not wt:
                c0, c1 = pair
                road = neighbor[0]
                dist = methodfunc(road.distance, road.speed,
                                  road.speedcode, road.units,
                                  units, codetable, prefertable,
                                  defaultprefer)
                wt = weightings[pair] = dist, road, neighbor[1]
                weightings[(c1,c0)] = dist, road, 1-neighbor[1]

            Q.insert(wt[0], pair)
            seen[pair] = 1

        while len(Q):
            (priority, (left, right)) = Q.pop()
            if priority < costs.get(right, MAXDIST):
                reverse_path[right] = left
                costs[right] = priority
                if right == end:
                    # We are done
                    path = []
                    this = right
                    pair = (right, reverse_path[right])
                    while this != start:
                        if this != end: path.append(weightings[pair][1:])
                        pair = (this, reverse_path[this])
                        this = pair[1]

                    path.append(weightings[pair][1:])
                    path.reverse()
                    return path

                for neighbor in right.roads:
                    pair = (right, neighbor[2])
                    if not seen.has_key(pair):
                        wt = weightings.get(pair)
                        if not wt:
                            c0, c1 = pair
                            road = neighbor[0]
                            dist = methodfunc(road.distance, road.speed,
                                              road.speedcode, road.units,
                                              units, codetable, prefertable,
                                              defaultprefer)
                            wt = weightings[pair] = dist, road, neighbor[1]
                            weightings[(c1,c0)] = dist, road, 1-neighbor[1]
                        Q.insert(wt[0]+priority, pair)
                        seen[pair] = 1
        
        raise ValueError, "start unreachable from end"

# End of RPDatabase
